package	jp.co.sds_corp.fixconnect.transport;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.Date;
import java.util.Vector;

import jp.co.sds_corp.fixconnect.FIXException;
import jp.co.sds_corp.fixconnect.message.FIXMessage;
import jp.co.sds_corp.fixconnect.message.FIXParser;
import jp.co.sds_corp.fixconnect.message.IMessageReceiver;
import jp.co.sds_corp.fixconnect.message.MessageReceiveEvent;
import jp.co.sds_corp.util.common.CmnLog;

public class FIXSocket
{
	Socket	sock;
	OutputStream	sout;
	InputStream		sin;
	Thread	thr;
	byte[] recvBuf;
	ByteArrayOutputStream bos;

	Vector<ITransportEventListener>	vecTrn = new Vector<ITransportEventListener>();
	Vector<IMessageReceiver>	vecRcv = new Vector<IMessageReceiver>();

	public FIXSocket() {
		bos = new ByteArrayOutputStream();
		socketTimeout = 1000;
		recvBufSize = 10240;
		closed = true;
	}

	public FIXSocket( String host, int port ) {
		this();
		this.host = host;
		this.port = port;
	}

	public FIXSocket( Socket sock ) {
		this();
		this.sock = sock;
		closed = false;
	}

	public void start() throws Exception {
		if ( sock == null ) {
			sock = new Socket( host, port );
			closed = false;
		}
		fireEventOccured( TransportEvent.Type.CONNECT );

		recvBuf = new byte[recvBufSize];
		sin = sock.getInputStream();
		sout = sock.getOutputStream();
		sock.setSoTimeout( getSocketTimeout() );

		thr = new Thread( new Receiver() );
		thr.start();
	}

	public void close() throws Exception {
		setClosed( true );
		if ( thr != null ) {
			CmnLog.info( this, "closing Socket..." );
			thr.interrupt();
			try {
				thr.join();
			} catch ( Exception ex ) {
			}
		}
		sock.close();
		CmnLog.info( this, "socket closed. " + toString() );
	}

	public void addEventListener( ITransportEventListener l ) {
		vecTrn.add( l );
	}
	public boolean removeEventListener( ITransportEventListener l ) {
		return vecTrn.remove( l );
	}
	public void fireEventOccured( TransportEvent ev ) {
		for ( int i = 0; i < vecTrn.size(); i++ ) {
			ITransportEventListener l = vecTrn.get( i );
			l.onEvent( ev );
		}
	}
	public void fireEventOccured( TransportEvent.Type type ) {
		TransportEvent ev = new TransportEvent( this, type );
		fireEventOccured( ev );
	}
	public void fireExceptionOccured( FIXException ex ) {
		for ( int i = 0; i < vecTrn.size(); i++ ) {
			ITransportEventListener l = vecTrn.get( i );
			l.onException( ex );
		}
	}

	public void addMessageReceiver( IMessageReceiver l ) {
		vecRcv.add( l );
	}
	public boolean removeMessageReceiver( IMessageReceiver l ) {
		return vecRcv.remove( l );
	}
	public void fireMessageReceived( FIXMessage msg ) {
		for ( int i = 0; i < vecRcv.size(); i++ ) {
			IMessageReceiver l = vecRcv.get( i );
			MessageReceiveEvent ev = new MessageReceiveEvent( this, msg );
			l.onMessage( ev );
		}
	}

	public synchronized void send( FIXMessage msg ) throws Exception {
		byte[] buf = parser.getBytes( msg );
		msg.setBuffer( buf );
		sout.write( buf );
	}

	public FIXMessage receive( int timeout ) throws Exception {
		long lastReceived = new Date().getTime();
		for ( ;; ) {
			if ( bos.size() > 0 ) {
				byte[] buf = bos.toByteArray();
				FIXMessage msg = null;
				try {
					msg = parser.parse( buf, 0, buf.length );
				} catch ( FIXException ex ) {
					CmnLog.warn( this, "parse Exception ", ex );
					// @@@ 使用済みバッファの消去
					// どうする？
				}
				if ( msg != null ) {
					byte[] msgBuf = new byte[msg.getLength()];
					System.arraycopy( buf, 0, msgBuf, 0, msg.getLength() );
					msg.setBuffer( msgBuf );
					// parse済みバッファの消去
					bos.reset();
					if ( buf.length > msg.getLength()) {
						bos.write( buf, msg.getLength(),
									buf.length - msg.getLength() );
					}
					return msg;
				}
			}

			int nRecv = -1;
			try {
				nRecv = sin.read( recvBuf, 0, recvBufSize );
			} catch ( SocketTimeoutException sex ) {
				if ( isClosed()) {
					return null;
				}
				long now = new Date().getTime();
				if ( timeout > 0 && ( now - lastReceived ) > timeout ) {
					return null;
				}
				continue;
			}
			if ( nRecv < 0 ) {
				throw( new SocketException( "connection closed" ));
			}
			lastReceived = new Date().getTime();
			bos.write( recvBuf, 0, nRecv );
		}
	}

	class Receiver implements Runnable
	{
		public void run() {
			for ( ;; ) {
				try {
					FIXMessage msg = receive( 0 );
					if ( msg != null ) {
						fireMessageReceived( msg );
					}
					if ( isClosed()) {
						fireEventOccured( TransportEvent.Type.CLOSE );
						break;
					}
				} catch ( Exception ex ) {
					fireEventOccured( TransportEvent.Type.CLOSE );
					FIXException fex = new FIXException( "receive error", ex );
					fireExceptionOccured( fex );
					break;
				}
			}
			if ( !isClosed()) {
				fireEventOccured( TransportEvent.Type.CLOSE );
				try {
					sock.close();
				} catch ( Exception ex ) {
				}
			}
		}
	}

	public Socket getSocket() {
		return sock;
	}

	public InetSocketAddress getRemoteAddress() {
		SocketAddress addr = sock.getRemoteSocketAddress();
		InetSocketAddress iaddr = (InetSocketAddress )addr;
		return iaddr;
	}

	public String toString() {
		String ret = "FIXSocket";
		InetSocketAddress addr = getRemoteAddress();
		if ( addr != null ) {
			ret += " remote=[" + addr + "]";
		}
		ret += " recvBufSize=" + getRecvBufSize();
		ret += " socketTimeout=" + getSocketTimeout();
		if ( isDebug()) {
			ret +=" DEBUG";
		}
		if ( isClosed()) {
			ret +=" CLOSED";
		}
		return ret;
	}

	String	host;	// 接続先host
	int	port;	// 接続先port
	boolean	closed;	// Socketクローズフラグ
	FIXParser	parser;	// メッセージのパーサ
	int	recvBufSize;	// 受信バッファサイズ
	int	socketTimeout;	// Socketタイムアウト
	boolean	debug;	// debugフラグ

	/**
	* 接続先hostを設定します。
	* @param	var		設定するデータ
	*/
	public void setHost( String var ) {
		this.host = var;
	}
	/**
	* 接続先hostを取得します。
	* @return	接続先host
	*/
	public String getHost() {
		return host;
	}

	/**
	* 接続先portを設定します。
	* @param	var		設定するデータ
	*/
	public void setPort( int var ) {
		this.port = var;
	}
	/**
	* 接続先portを取得します。
	* @return	接続先port
	*/
	public int getPort() {
		return port;
	}

	/**
	* Socketクローズフラグを設定します。
	* @param	var		設定するデータ
	*/
	public void setClosed( boolean var ) {
		this.closed = var;
	}
	/**
	* Socketクローズフラグを取得します。
	* @return	Socketクローズフラグ
	*/
	public boolean isClosed() {
		return closed;
	}

	/**
	* メッセージのパーサを設定します。
	* @param	var		設定するデータ
	*/
	public void setParser( FIXParser var ) {
		this.parser = var;
	}
	/**
	* メッセージのパーサを取得します。
	* @return	メッセージのパーサ
	*/
	public FIXParser getParser() {
		return parser;
	}

	/**
	* 受信バッファサイズを設定します。
	* @param	var		設定するデータ
	*/
	public void setRecvBufSize( int var ) {
		this.recvBufSize = var;
	}
	/**
	* 受信バッファサイズを取得します。
	* @return	受信バッファサイズ
	*/
	public int getRecvBufSize() {
		return recvBufSize;
	}

	/**
	* Socketタイムアウトを設定します。
	* @param	var		設定するデータ
	*/
	public void setSocketTimeout( int var ) {
		this.socketTimeout = var;
	}
	/**
	* Socketタイムアウトを取得します。
	* @return	Socketタイムアウト
	*/
	public int getSocketTimeout() {
		return socketTimeout;
	}

	/**
	* debugフラグを設定します。
	* @param	var		設定するデータ
	*/
	public void setDebug( boolean var ) {
		this.debug = var;
	}
	/**
	* debugフラグを取得します。
	* @return	debugフラグ
	*/
	public boolean isDebug() {
		return debug;
	}

}
